package implementation;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Queue;
import java.util.Scanner;
import java.util.Stack;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

public class PuzzleSolver {

    public static final int POOLSIZE = 4;

    // Share these Attributes
    private static int[] keys;
    private static int dimX, dimY, dimZ;
    private static boolean[][][] matchingCache;
    private static ThreadPoolExecutor threadServices;
    private static ArrayList<Integer> cornerTypes = new ArrayList<>(), edgeTypes = new ArrayList<>(),
            faceTypes = new ArrayList<>(), innerTypes = new ArrayList<>();
    private static Queue<Future<Boolean>> results;

    public static boolean innerSolution;

    // This is the solution
    private static int[][][] solution;
    private static HashMap<Integer, QuaderType> sStateToType;

    private static boolean hasSolution;

    // Need to copy these Attributes
    private HashMap<Integer, QuaderType> stateToType;
    private int[][][] grid; // Array of stateNumbers
    private static ArrayDeque<PPoint> innerPath, outerPath;
    private ArrayDeque<PPoint> path;

    Stack<PositionQueue> workingStates;

    public PuzzleSolver(int x, int y, int z) {
        dimX = x;
        dimY = y;
        dimZ = z;

        grid = new int[x][y][z];

        stateToType = new HashMap<>();
    }

    public PuzzleSolver(int[][][] grid, HashMap<Integer, QuaderType> stateToType,
            PositionQueue workingState, ArrayDeque<PPoint> path) {

        // Copy the Grid
        this.grid = new int[dimX][dimY][dimZ];
        for (int x = 0; x < dimX; x++) {
            for (int y = 0; y < dimY; y++) {
                for (int z = 0; z < dimZ; z++) {
                    this.grid[x][y][z] = grid[x][y][z];
                }
            }
        }

        // Clone the HashMap and ArrayList
        this.stateToType = new HashMap<>();
        ArrayList<QuaderType> quaderTypes = new ArrayList<>();

        for (int state : stateToType.keySet()) {
            QuaderType type = stateToType.get(state);

            int index = quaderTypes.indexOf(type);
            if (index == -1) {
                type = type.clone();
                quaderTypes.add(type);
                this.stateToType.put(state, type);
            } else {
                type = quaderTypes.get(index);
                this.stateToType.put(state, type);
            }
        }

        // Clone the working stack
        this.workingStates = new Stack<>();
        this.workingStates.push(workingState.clone());

        // Copy the Path through the Array
        this.path = path.clone();
    }

    private void init() {

        // Get all StateNumbers as keys of the HashMap
        keys = new ArrayList<>(stateToType.keySet()).stream().mapToInt(a -> a).toArray();

        // Compute the matching cache
        matchingCache = calculateMatchingCache();

        // Save current State in Stack of states
        workingStates = new Stack<>();

        // Init the Excecuter-Service
        threadServices = (ThreadPoolExecutor) Executors.newFixedThreadPool(POOLSIZE);

        // At first, there is no Solution
        hasSolution = false;

        innerSolution = !(dimX >= 3 && dimY >= 3 && dimZ >= 3);

        // Init the Futures-Queue
        results = new ArrayDeque<>();

        solution = new int[dimX][dimY][dimZ];

        sStateToType = stateToType;
    }

    public boolean solve() {

        init();

        // Calculate Path through the Array
        innerPath = getInnerPathThroughArray(dimX, dimY, dimZ);
        outerPath = getOuterPathThroughArray(dimX, dimY, dimZ, innerPath);
        path = new ArrayDeque<>(innerPath);
        path.addAll(outerPath);

        if (outerPath.size() > 0) {
            PPoint firstPoint = path.poll();
            PositionQueue first = new PositionQueue(firstPoint, getPossibleStateNumbers(firstPoint, solution));

            PuzzleSolver solver = new PuzzleSolver(solution, stateToType, first, path);

            // In den Threadpool hinzufügen:
            try {
                results.offer(threadServices.submit(solver::execute));

                Future<Boolean> result;
                while ((result = results.poll()) != null && !hasSolution) {
                    result.get();
                }

                Thread.sleep(100);

                threadServices.shutdownNow();

                return threadServices.awaitTermination(2, TimeUnit.SECONDS) && hasSolution;
            } catch (InterruptedException | ExecutionException e) {
                e.printStackTrace();
            }
        }

        return false;
    }

    private long count = 0, sum = 0;

    private boolean execute() {
        while (!workingStates.isEmpty() && !hasSolution) {

            if (count % 50000 == 0) {
                // System.out.println(path.size());
            }

            PositionQueue queue = workingStates.peek();

            if (!queue.isEmpty()) {

                // Altes Teil herausnehmen
                if (queue.isSet()) {
                    stateToType.get(grid[queue.getX()][queue.getY()][queue.getZ()]).addOccurence();
                    queue.setSet(false);
                }

                // neuen Thread erzeugen, wenn noch verfügbar
                int minPathSize = 10;

                if (queue.size() > 1 && path.size() <= dimX * dimY * dimZ - innerPath.size()
                        && path.size() > minPathSize && threadServices.getActiveCount() < POOLSIZE) {
                    count = 0;

                    PuzzleSolver array = new PuzzleSolver(grid, stateToType, workingStates.peek(), path);

                    array.workingStates.peek().clearAndAdd(queue.poll());

                    try {

                        results.offer(threadServices.submit(array::execute));
                    } catch (Exception e) {
                        return false;
                    }
                }

                count++;

                // Neues Teil hineinsetzen
                grid[queue.getX()][queue.getY()][queue.getZ()] = queue.poll();
                queue.setSet(true);
                stateToType.get(grid[queue.getX()][queue.getY()][queue.getZ()]).subtractOccurance();

                // Abbruchbedingung
                if (path.isEmpty()) {
                    hasSolution = true;
                    System.out.println("Success :)");
                    solution = this.grid;
                    sStateToType = stateToType;

                    return true;
                }

                // Neuen Punkt besuchen:
                PPoint p = path.pop();

                PositionQueue newQueue = new PositionQueue(p, getPossibleStateNumbers(p, grid));

                if (newQueue.size() > 0) {
                    workingStates.push(newQueue);
                } else {
                    // Besuchten Punkt wieder freigeben!
                    path.push(p);
                }
            } else {
                // Teil herausnehmen
                stateToType.get(grid[queue.getX()][queue.getY()][queue.getZ()]).addOccurence();
                grid[queue.getX()][queue.getY()][queue.getZ()] = 0;

                workingStates.pop();

                // Besuchten Punkt wieder freigeben!
                path.push(new PPoint(queue.getX(), queue.getY(), queue.getZ()));
            }

        }

        return false;
    }

    private static ArrayDeque<PPoint> getInnerPathThroughArray(int dimX, int dimY, int dimZ) {
        ArrayDeque<PPoint> result = new ArrayDeque<>();

        int distance = 0;

        // Erst nur innen
        while (distance <= dimX + dimY + dimZ) {
            for (int x = 1; x < dimX - 1; x++) {
                for (int y = 1; y < dimY - 1; y++) {
                    for (int z = 1; z < dimZ - 1; z++) {
                        PPoint p = new PPoint(x, y, z);
                        if (!result.contains(p)
                                && Math.abs(dimX / 2 - p.getX()) + Math.abs(dimY / 2 - p.getY())
                                        + Math.abs(dimZ / 2 - p.getZ()) == distance) {
                            result.addLast(p);
                        }
                    }
                }
            }

            ++distance;
        }

        return result;
    }

    private static ArrayDeque<PPoint> getOuterPathThroughArray(int dimX, int dimY, int dimZ,
            ArrayDeque<PPoint> innerPath) {
        ArrayDeque<PPoint> result = new ArrayDeque<>();

        int distance = 0;

        // Dann außen
        while (distance <= dimX + dimY + dimZ) {
            for (int x = 0; x < dimX; x++) {
                for (int y = 0; y < dimY; y++) {
                    for (int z = 0; z < dimZ; z++) {
                        PPoint p = new PPoint(x, y, z);
                        if (!innerPath.contains(p) && x + y + z == distance) {
                            result.addLast(p);
                        }
                    }
                }
            }

            ++distance;
        }

        return result;
    }

    private ArrayDeque<Integer> getPossibleStateNumbers(PPoint p, int[][][] grid) {
        ArrayDeque<Integer> result = new ArrayDeque<>();
        ArrayList<Integer> keys = null;
        switch (Quader.getType(p.getX(), p.getY(), p.getZ(), dimX, dimY, dimZ)) {
            case INNER -> {
                keys = innerTypes;
            }
            case FACE -> {
                keys = faceTypes;
            }
            case EDGE -> {
                keys = edgeTypes;
            }
            case CORNER -> {
                keys = cornerTypes;
            }
        }

        for (int i = 0; i < keys.size(); ++i) {
            int stateNumber = keys.get(i);
            if (stateToType.get(stateNumber).getNumberOfOccurrences() > 0 && fitsInPlace(stateNumber, p, grid)) {
                result.offer(stateNumber);
            }
        }

        return result;
    }

    private static boolean fitsInPlace(int stateNumber, PPoint p, int[][][] grid) {
        boolean fits = true;

        int x = p.getX(), y = p.getY(), z = p.getZ();

        int needZeros = 0;
        int countZeros = 0;

        if (x == dimX - 1) {
            needZeros++;
        }
        if (x == 0) {
            needZeros++;
        }
        if (y == dimY - 1) {
            needZeros++;
        }
        if (y == 0) {
            needZeros++;
        }
        if (z == dimZ - 1) {
            needZeros++;
        }
        if (z == 0) {
            needZeros++;
        }

        // TOP
        int faceNumber = stateNumber / 100000;
        if (faceNumber == 0) {
            countZeros++;
        }

        if (z == dimZ - 1) {
            fits &= matchingCache[1][faceNumber][5];
        } else {
            fits &= matchingCache[1][faceNumber][grid[x][y][z + 1] % 10];
        }

        // BOTTOM
        faceNumber = stateNumber % 10;
        if (faceNumber == 0) {
            countZeros++;
        }

        if (z == 0) {
            fits &= matchingCache[1][faceNumber][5];
        } else {
            fits &= matchingCache[1][faceNumber][grid[x][y][z - 1] / 100000];
        }

        // FRONT
        faceNumber = stateNumber / 1000 % 10;
        if (faceNumber == 0) {
            countZeros++;
        }

        if (x == dimX - 1) {
            fits &= matchingCache[0][faceNumber][5];
        } else {
            fits &= matchingCache[0][faceNumber][grid[x + 1][y][z] / 10 % 10];
        }

        // Back
        faceNumber = stateNumber / 10 % 10;
        if (faceNumber == 0) {
            countZeros++;
        }

        if (x == 0) {
            fits &= matchingCache[0][faceNumber][5];
        } else {
            fits &= matchingCache[0][faceNumber][grid[x - 1][y][z] / 1000 % 10];
        }

        // RIGHT
        faceNumber = stateNumber / 100 % 10;
        if (faceNumber == 0) {
            countZeros++;
        }

        if (y == dimY - 1) {
            fits &= matchingCache[0][faceNumber][5];
        } else {
            fits &= matchingCache[0][faceNumber][grid[x][y + 1][z] / 10000 % 10];
        }

        // LEFT
        faceNumber = stateNumber / 10000 % 10;
        if (faceNumber == 0) {
            countZeros++;
        }

        if (y == 0) {
            fits &= matchingCache[0][faceNumber][5];
        } else {
            fits &= matchingCache[0][faceNumber][grid[x][y - 1][z] / 100 % 10];
        }

        return fits && countZeros == needZeros;
    }

    private static boolean fits(boolean top, int objectFace, int satelliteFace) {
        if (top) {
            if (objectFace != 0 && satelliteFace != 0 && objectFace + satelliteFace == 5) {
                return true;
            } else if (objectFace != 0 && satelliteFace == 0) {
                return true;
            }
        } else {
            int test = objectFace + satelliteFace;
            if (objectFace != 0 && satelliteFace != 0 && (test == 3 || test == 7)) {
                return true;
            } else if (objectFace != 0 && satelliteFace == 0) {
                return true;
            }
        }

        return false;
    }

    public void add(Quader quader, int partNumber) {
        var stateNumbers = quader.getAllStateNumbers(dimX, dimY, dimZ);

        if (stateNumbers.size() != 0) {
            if (stateToType.containsKey(stateNumbers.get(0))) {
                var type = stateToType.get(stateNumbers.get(0));
                type.addOccurence();
                type.addPartNumber(partNumber);
            } else {
                QuaderType newType = new QuaderType(quader, null, 1);
                newType.addPartNumber(partNumber);

                for (int stateNumber : stateNumbers) {
                    stateToType.put(stateNumber, newType);
                }

                switch (quader.getType(dimX, dimY, dimZ)) {
                    case CORNER -> {
                        for (int stateNumber : stateNumbers) {
                            if (!cornerTypes.contains(stateNumber)) {
                                cornerTypes.add(stateNumber);
                            }
                        }
                    }
                    case EDGE -> {
                        for (int stateNumber : stateNumbers) {
                            if (!edgeTypes.contains(stateNumber)) {
                                edgeTypes.add(stateNumber);
                            }
                        }
                    }
                    case FACE -> {
                        for (int stateNumber : stateNumbers) {
                            if (!faceTypes.contains(stateNumber)) {
                                faceTypes.add(stateNumber);
                            }
                        }
                    }
                    case INNER -> {
                        for (int stateNumber : stateNumbers) {
                            if (!innerTypes.contains(stateNumber)) {
                                innerTypes.add(stateNumber);
                            }
                        }
                    }
                }
            }
        }
    }

    public int[][][] getSolution() {
        return solution;
    }

    public HashMap<Integer, QuaderType> getStateToType() {
        return stateToType;
    }

    public static boolean[][][] calculateMatchingCache() {
        // Direction faceOfObject faceOfSatellite
        boolean[][][] matches = new boolean[2][5][6];

        for (int i = 0; i < 2; ++i) {
            for (int o = 0; o < 5; ++o) {
                for (int s = 0; s < 6; ++s) {
                    if (s == 5) {
                        matches[i][o][s] = o == 0;
                    } else {
                        if (i == 0) {
                            matches[i][o][s] = fits(false, o, s);
                        } else {
                            matches[i][o][s] = fits(true, o, s);
                        }
                    }
                }
            }
        }

        return matches;
    }

    public void writeSolutionTo(File file) {
        System.out.println(file.getAbsolutePath());
        try (PrintWriter pw = new PrintWriter(file)) {
            var solution = getSolution();
            var stateToType = getStateToType();

            int dimX = solution.length;
            int dimY = solution[0].length;
            int dimZ = solution[0][0].length;

            pw.printf("Dimension: %d,%d,%d\n", dimX, dimY, dimZ);

            for (int x = 0; x < solution.length; ++x) {
                for (int y = 0; y < solution[x].length; ++y) {
                    for (int z = 0; z < solution[x][y].length; ++z) {
                        int stateNumber = solution[x][y][z];
                        int partNumber = stateToType.get(stateNumber).getPartNumbers().remove(0);

                        pw.printf("[%d,%d,%d] Teil %d: %s\n", x + 1, y + 1, z + 1, partNumber,
                                Quader.getStateString(stateNumber));
                    }
                }
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
    }

    public static PuzzleSolver parseInput(File file) {
        PuzzleSolver result = null;

        try {
            Scanner sc = new Scanner(file);
            while (sc.hasNext()) {
                String currentLine = sc.nextLine();
                if (currentLine.startsWith("//")) {
                    continue;
                }

                if (currentLine.startsWith("Dimension")) {
                    String[] dimensions = currentLine.split(" |,");
                    int dimX = Integer.parseInt(dimensions[1]);
                    int dimY = Integer.parseInt(dimensions[2]);
                    int dimZ = Integer.parseInt(dimensions[3]);

                    int[] dim = new int[] { dimX, dimY, dimZ };
                    Arrays.sort(dim);

                    result = new PuzzleSolver(dim[2], dim[1], dim[0]);
                }

                if (currentLine.startsWith("Teil") && result != null) {
                    String[] sQuader = currentLine.split(" ");

                    int[] quader = new int[6];
                    quader[0] = Integer.parseInt(sQuader[2]);
                    quader[1] = Integer.parseInt(sQuader[3]);
                    quader[2] = Integer.parseInt(sQuader[4]);
                    quader[3] = Integer.parseInt(sQuader[5]);
                    quader[4] = Integer.parseInt(sQuader[6]);
                    quader[5] = Integer.parseInt(sQuader[7]);

                    result.add(new Quader(quader), Integer.parseInt(sQuader[1].split(":")[0]));
                }
            }

            sc.close();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }

        return result;
    }
}
