package implementation;

import java.util.ArrayList;
import java.util.Arrays;

public class Quader {

    public static enum TYPE {
        CORNER, EDGE, FACE, INNER
    }

    public static enum Axis {
        X, Y, Z
    }

    private int[] faces;
    private TYPE type;

    public Quader(int[] faces) {
        this.faces = faces;

        int countZeros = 0;
        for (int i = 0; i < faces.length; i++) {
            if (faces[i] == 0) {
                countZeros++;
            }
        }

        switch (countZeros) {
            case 0 -> this.type = TYPE.INNER;
            case 1 -> this.type = TYPE.FACE;
            case 2 -> this.type = TYPE.EDGE;
            default -> this.type = TYPE.CORNER;
        }
    }

    public Quader clone() {
        return new Quader(faces.clone());
    }

    public boolean equals(Object obj) {
        if (obj instanceof Quader) {
            Quader q = (Quader) obj;
            for (int i = 0; i < faces.length; ++i) {
                if (faces[i] != q.faces[i]) {
                    return false;
                }
            }
            return true;
        } else {
            return false;
        }
    }

    public void rotateQuader(Axis a, int n) {
        switch (a) {
            case X -> {
                for (int i = 0; i < n; ++i) {
                    rotateQuaderX();
                }
            }
            case Y -> {
                for (int i = 0; i < n; ++i) {
                    rotateQuaderY();
                }
            }
            case Z -> {
                for (int i = 0; i < n; ++i) {
                    rotateQuaderZ();
                }
            }
        }
    }

    private void rotateQuaderX() {
        int temp1 = faces[1];

        faces[1] = rotateFace(faces[0], false);
        faces[0] = rotateFace(faces[3], false);
        faces[3] = rotateFace(faces[5], false);
        faces[5] = rotateFace(temp1, false);

        faces[2] = rotateFace(faces[2], false);
        faces[4] = rotateFace(faces[4], true);
    }

    private void rotateQuaderY() {
        int temp5 = faces[5];

        faces[5] = faces[2];
        faces[2] = faces[0];
        faces[0] = rotateFace(rotateFace(faces[4], true), true);
        faces[4] = rotateFace(rotateFace(temp5, false), false);

        faces[1] = rotateFace(faces[1], true);
        faces[3] = rotateFace(faces[3], false);
    }

    private void rotateQuaderZ() {
        int temp4 = faces[4];

        faces[4] = faces[3];
        faces[3] = faces[2];
        faces[2] = faces[1];
        faces[1] = temp4;

        faces[0] = rotateFace(faces[0], false);
        faces[5] = rotateFace(faces[5], true);
    }

    private static int rotateFace(int before, boolean right) {
        if (before == 0) {
            return 0;
        }

        if (right) {
            return before % 4 + 1;
        } else {
            before -= 1;
            return before == 0 ? 4 : before;
        }
    }

    public String toString() {
        return Arrays.toString(faces);
    }

    private int getStateNumber() {
        return faces[5] + faces[4] * 10 + faces[3] * 100 + faces[2] * 1000 + faces[1] * 10000 + faces[0] * 100000;
    }

    public ArrayList<Integer> getAllStateNumbers(int dimX, int dimY, int dimZ) {
        ArrayList<Integer> stateNumbers = new ArrayList<>();

        for (int x = 0; x < 4; ++x) {
            this.rotateQuader(Axis.X, 1);
            for (int z = 0; z < 4; ++z) {
                this.rotateQuader(Axis.Z, 1);

                int number = this.getStateNumber();
                if (!stateNumbers.contains(number) && fitsInField(this.faces, dimX, dimY, dimZ)) {
                    stateNumbers.add(number);
                }
            }
        }

        this.rotateQuader(Axis.Y, 1);

        for (int z = 0; z < 4; ++z) {
            this.rotateQuader(Axis.Z, 1);

            int number = this.getStateNumber();
            if (!stateNumbers.contains(number) && fitsInField(this.faces, dimX, dimY, dimZ)) {
                stateNumbers.add(number);
            }
        }

        this.rotateQuader(Axis.Y, 2);

        for (int z = 0; z < 4; ++z) {
            this.rotateQuader(Axis.Z, 1);

            int number = this.getStateNumber();
            if (!stateNumbers.contains(number) && fitsInField(this.faces, dimX, dimY, dimZ)) {
                stateNumbers.add(number);
            }
        }

        return stateNumbers;
    }

    public static boolean fitsInField(int[] state, int dimX, int dimY, int dimZ) {
        boolean fits = true;
        if (dimX == 1) {
            fits &= state[2] == 0 && state[4] == 0;
        } else if (dimX == 2) {
            fits &= state[2] == 0 ^ state[4] == 0;
        }

        if (dimY == 1) {
            fits &= state[1] == 0 && state[3] == 0;
        } else if (dimY == 2) {
            fits &= state[1] == 0 ^ state[3] == 0;
        }

        if (dimZ == 1) {
            fits &= state[0] == 0 && state[5] == 0;
        } else if (dimZ == 2) {
            fits &= state[0] == 0 ^ state[5] == 0;
        }

        return fits;
    }

    public static int[] getStateArray(int stateNumber) {
        int[] quader = new int[6];

        for (int i = 5; i >= 0; --i) {
            quader[i] = stateNumber % 10;
            stateNumber /= 10;
        }

        return quader;
    }

    public TYPE getType(int dimX, int dimY, int dimZ) {
        return this.type;
    }

    public static TYPE getType(int x, int y, int z, int dimX, int dimY, int dimZ) {
        boolean xOuter = (x == 0 || x == dimX - 1);
        boolean yOuter = (y == 0 || y == dimY - 1);
        boolean zOuter = (z == 0 || z == dimZ - 1);

        if (xOuter && yOuter && zOuter) {
            return TYPE.CORNER;
        } else if ((xOuter && yOuter) || (xOuter && zOuter) || (yOuter && zOuter)) {
            if (Math.min(Math.min(dimX, dimY), dimZ) == 1) {
                return TYPE.CORNER;
            }
            return TYPE.EDGE;
        } else if ((xOuter && !yOuter && !zOuter) || (!xOuter && yOuter && !zOuter) || (!xOuter && !yOuter && zOuter)) {
            if (Math.min(Math.min(dimX, dimY), dimZ) == 1) {
                return TYPE.EDGE;
            } else if (((dimX == dimY) || (dimY == dimZ) || (dimX == dimY)) && (dimX == 1 || dimZ == 1)) {
                return TYPE.CORNER;
            }
            return TYPE.FACE;
        } else {
            return TYPE.INNER;
        }
    }

    public static String getStateString(int stateNumber) {
        int[] quader = new int[6];

        for (int i = 5; i >= 0; --i) {
            quader[i] = stateNumber % 10;
            stateNumber /= 10;
        }

        boolean first = true;
        String result = "";
        for (int i = 0; i < quader.length; ++i) {
            if (first) {
                first = false;
            } else {
                result += " ";
            }

            result += quader[i];
        }

        return result;
    }
}
